"""Known attributes for reference purposes only."""
KNOWN_ATTRIBUTES = sorted("""ERRORLOG;ENTRIES;MAX
GROUP;LIST
LEARN;MINSPEED
FAN;DIR
NW;PARAMS;ACTUAL
SMARTMODE;STATE
FW;NAME
LEARN;ZEROTEMP
WINTERMODE;STATE
DEVICE;SERVER
SMARTSLEEP;MINSPEED
LIGHT;LEVEL;ACTUAL
LEARN;MAXSPEED
ERRORLOG;ENTRIES;NUM
DEVICE;LIGHT
FW;FW000007
LIGHT;PWR
SLEEP;EVENT
GROUP;ROOM;TYPE
NW;DHCP
DEVICE;INDICATORS
SNSROCC;TIMEOUT;MAX
NW;AP;STATUS
FAN;AUTO
SMARTMODE;ACTUAL
FAN;SPD;MIN
NW;TOKEN
DEVICE;BEEPER
TIME;VALUE
LIGHT;LEVEL;MIN
LIGHT;AUTO
SMARTSLEEP;IDEALTEMP
SLEEP;STATE
FAN;WHOOSH;STATUS
SNSROCC;STATUS
SNSROCC;TIMEOUT;CURR
NAME;VALUE
NW;SSID
LIGHT;BOOKENDS
FAN;TIMER;CURR
FAN;TIMER;MAX
FAN;SPD;ACTUAL
FAN;SPD;MAX
FAN;PWR
WINTERMODE;HEIGHT
SCHEDULE;CAP
LIGHT;LEVEL;MAX
SNSROCC;TIMEOUT;MIN
FAN;BOOKENDS
FAN;TIMER;MIN
SLEEP;EVENT;OFF
SMARTSLEEP;MAXSPEED
SCHEDULE;EVENT;LIST""".split('\n'))
